from dataclasses import asdict
from typing import Any, Callable, Iterable, Type

from pymongo.collection import Collection

from mongorepo.utils import convert_to_dto
from mongorepo import DTO, exceptions


def _add_method(dto_type: Type[DTO], collection: Collection) -> Callable:
    def add(self, dto: DTO) -> DTO:
        collection.insert_one(asdict(dto))
        return dto
    return add


def _get_all_method(dto_type: Type[DTO], collection: Collection) -> Callable:
    def get_all(self, **filters: Any) -> Iterable[DTO]:
        cursor = collection.find(filters)
        for dct in cursor:
            yield convert_to_dto(dto_type, dct)
    return get_all


def _update_method(dto_type: Type[DTO], collection: Collection) -> Callable:
    def update(self, dto: DTO, **filters: Any) -> DTO:
        data: dict[str, dict[str, Any]] = {'$set': {}}
        for field, value in asdict(dto).items():
            if isinstance(value, (int, bool, float)):
                data['$set'][field] = value
            elif not value:
                continue
            else:
                data['$set'][field] = value
        collection.find_one_and_update(filter=filters, update=data, return_document=True)
        return dto
    return update


def _delete_method(dto_type: Type[DTO], collection: Collection) -> Callable:
    def delete(self, **filters: Any) -> bool:
        deleted = collection.find_one_and_delete(filters)
        return True if deleted else False
    return delete


def _get_method(dto_type: Type[DTO], collection: Collection) -> Callable:
    def get(self, **filters: Any) -> DTO | None:
        result = collection.find_one(filters)
        return convert_to_dto(dto_type, result) if result else None
    return get


def _update_field_method(dto_type: Type[DTO], collection: Collection) -> Callable:
    def update_field(self, field_name: str, value: Any, **filters) -> DTO | None:
        if field_name not in dto_type.__dict__['__annotations__']:
            raise exceptions.MongoRepoException(
                message=f'{dto_type} does not have field "{field_name}"'
            )
        result = collection.find_one_and_update(
            filter=filters, update={'$set': {field_name: value}}, return_document=True,
        )
        return convert_to_dto(dto_type, result) if result else None
    return update_field


def _update_integer_field_method(
    dto_type: Type[DTO], collection: Collection, field_name: str, _weight: int = 1,
) -> Callable:
    def update_interger_field(self, weight: int | None = None, **filters) -> DTO | None:
        w = weight if weight is not None else _weight
        document = collection.find_one_and_update(
            filter=filters, update={'$inc': {field_name: w}}, return_document=True
        )
        return convert_to_dto(dto_type=dto_type, dct=document) if document else None
    return update_interger_field


def _update_list_field_method(
    dto_type: Type[DTO], collection: Collection, field_name: str, command: str = '$push',
) -> Callable:
    def update_list(self, value: Any, **filters) -> DTO | None:
        document = collection.find_one_and_update(
            filter=filters, update={command: {field_name: value}}, return_document=True
        )
        return convert_to_dto(dto_type=dto_type, dct=document) if document else None
    return update_list


def _pop_list_method(dto_type: Type[DTO], collection: Collection, field_name: str) -> Callable:
    def pop_list(self, **filters) -> Any | None:
        document = collection.find_one_and_update(
            filter=filters, update={'$pop': {field_name: 1}},
        )
        return document[field_name][-1] if document else None
    return pop_list


METHOD_NAME__CALLABLE: dict[str, Callable] = {
    'get': _get_method,
    'add': _add_method,
    'update': _update_method,
    'delete': _delete_method,
    'get_all': _get_all_method,
    'update_field': _update_field_method,
    'update_list': _update_list_field_method,
    'pop': _pop_list_method,
    'update_integer': _update_integer_field_method,
}



from typing import Callable, Annotated

from mongorepo.base import Access
from mongorepo.subst import substitute_class_methods
from mongorepo.handlers import _handle_cls


def mongo_repository(
    cls: type | None = None,
    add: bool = True,
    get: bool = True,
    get_all: bool = True,
    update: bool = True,
    delete: bool = True,
    update_field: bool = False,
    integer_fields: list[Annotated[str, 'field names']] | None = None,
    array_fields: list[Annotated[str, 'field names']] | None = None,
    method_access: Access | None = None,
) -> type | Callable:
    """
    ## Decorator for MongoDB repositories

    * decorated class must provide `Meta` class inside
    with variables "dto"(represent simple dataclass) and
    `collection` (represent mongo collection of type `Collection` from `pymongo` library)

    * You can also provide `index` field that can be just a string name of the field
    which you want to make index field or it can be instance of `mongorepo.Index`
    with extended settings

    * You can use `method_access` to make all methods
    private, protected or public (use `mongorepo.Access`)

    ### usage example:
    ```
    @mongo_repository(delete=False)
    class MongoRepository:
        class Meta:
            dto = UserDTO
            collection: Collection = db["users"]
            index = mongorepo.Index(field="name")
            method_access = mongorepo.Access.PROTECTED
    ```
    """

    def wrapper(cls) -> type:
        return _handle_cls(
            cls=cls,
            add=add,
            update=update,
            get_all=get_all,
            get=get,
            delete=delete,
            update_field=update_field,
            integer_fields=integer_fields,
            array_fields=array_fields,
            method_access=method_access,
        )

    if cls is None:
        return wrapper

    return wrapper(cls)


def substitute(source_cls: type) -> type | Callable:
    """
    ## !Not implemented

    * Decorator that can dynamically substitute methods of
      mongorepo decorators or classes with source class methods
    """
    def wrapper(target_cls) -> type:
        return substitute_class_methods(
            target_cls=target_cls,
            source_cls=source_cls
        )

    return wrapper



import asyncio
import pymongo
from mongorepo import Index
from motor.motor_asyncio import AsyncIOMotorCollection


def _run_asyncio_create_index(index: Index | str, collection: AsyncIOMotorCollection) -> None:
    """
    Tries to create an index for `AsyncIOMotorCollection` in async loop or after
    """
    loop = asyncio.get_running_loop()
    if loop.is_running():
        asyncio.create_task(_create_index_async(index=index, collection=collection))
    else:
        loop.run_until_complete(_create_index_async(index=index, collection=collection))


async def _create_index_async(index: Index | str, collection: AsyncIOMotorCollection) -> None:
    """
    ### Creates an index for the collection
    * index parameter can be string or mongorepo.Index
    * If index is string, create standard mongodb index
    * If it's `mongorepo.Index` creates index with user's settings
    """
    if isinstance(index, str):
        await collection.create_index(index)
        return
    index_name = f'index_{index.field}_1'
    if index.name:
        index_name = index.name
    direction = pymongo.DESCENDING if index.desc else pymongo.ASCENDING
    await collection.create_index(
        [(index.field, direction)],
        name=index_name,
        unique=index.unique,
    )



from dataclasses import asdict
from typing import Any, Callable, Type, AsyncGenerator

from motor.motor_asyncio import AsyncIOMotorCollection

from mongorepo.utils import convert_to_dto
from mongorepo import DTO, exceptions


def _add_method_async(dto_type: Type[DTO], collection: AsyncIOMotorCollection) -> Callable:
    async def add(self, dto: DTO) -> DTO:
        await collection.insert_one(asdict(dto))
        return dto
    return add


def _get_all_method_async(dto_type: Type[DTO], collection: AsyncIOMotorCollection) -> Callable:
    async def get_all(self, **filters: Any) -> AsyncGenerator[DTO, None]:
        cursor = collection.find(filters)
        async for dct in cursor:
            yield convert_to_dto(dto_type, dct)
    return get_all


def _update_method_async(dto_type: Type[DTO], collection: AsyncIOMotorCollection) -> Callable:
    async def update(self, dto: DTO, **filters: Any) -> DTO:
        data: dict[str, dict[str, Any]] = {'$set': {}}
        for field, value in asdict(dto).items():
            if isinstance(value, (int, bool, float)):
                data['$set'][field] = value
            elif not field:
                continue
            else:
                data['$set'][field] = value
        await collection.find_one_and_update(filter=filters, update=data)
        return dto
    return update


def _delete_method_async(dto_type: Type[DTO], collection: AsyncIOMotorCollection) -> Callable:
    async def delete(self, **filters: Any) -> bool:
        deleted = await collection.find_one_and_delete(filters)
        return True if deleted else False
    return delete


def _get_method_async(dto_type: Type[DTO], collection: AsyncIOMotorCollection) -> Callable:
    async def get(self, _id: str | None = None, **filters: Any) -> DTO | None:
        result = await collection.find_one(filters)
        return convert_to_dto(dto_type, result) if result else None
    return get


def _update_field_method_async(dto_type: Type[DTO], collection: AsyncIOMotorCollection) -> Callable:
    async def update_field(self, field_name: str, value: Any, **filters) -> DTO | None:
        if field_name not in dto_type.__dict__['__annotations__']:
            raise exceptions.MongoRepoException(
                message=f'{dto_type} does have field "{field_name}"'
            )
        result = await collection.find_one_and_update(
            filter=filters, update={'$set': {field_name: value}}, return_document=True,
        )
        return convert_to_dto(dto_type, result) if result else None
    return update_field


def _update_integer_field_method_async(
    dto_type: Type[DTO], collection: AsyncIOMotorCollection, field_name: str, _weight: int = 1,
) -> Callable:
    async def update_interger_field(self, weight: int | None = None, **filters) -> DTO | None:
        w = weight if weight is not None else _weight
        document = await collection.find_one_and_update(
            filter=filters, update={'$inc': {field_name: w}}, return_document=True
        )
        return convert_to_dto(dto_type=dto_type, dct=document) if document else None
    return update_interger_field


def _update_list_field_method_async(
    dto_type: Type[DTO],
    collection: AsyncIOMotorCollection,
    field_name: str,
    command: str = '$push',
) -> Callable:
    async def update_list(self, value: Any, **filters) -> Any:
        document = await collection.find_one_and_update(
            filter=filters, update={command: {field_name: value}}, return_document=True
        )
        return convert_to_dto(dto_type=dto_type, dct=document) if document else None
    return update_list


def _pop_list_method_async(
    dto_type: Type[DTO],
    collection: AsyncIOMotorCollection,
    field_name: str,
) -> Callable:
    async def pop_list(self, **filters) -> Any | None:
        document = await collection.find_one_and_update(
            filter=filters, update={'$pop': {field_name: 1}},
        )
        return document[field_name][-1] if document else None
    return pop_list


METHOD_NAME__CALLABLE: dict[str, Callable] = {
    'get': _get_method_async,
    'add': _add_method_async,
    'update': _update_method_async,
    'delete': _delete_method_async,
    'get_all': _get_all_method_async,
    'update_field': _update_field_method_async,
    'update_list': _update_list_field_method_async,
    'pop': _pop_list_method_async,
    'update_integer': _update_integer_field_method_async,
}



from dataclasses import is_dataclass
from typing import Any, Type, Union, get_args, get_origin

import pymongo
from pymongo.collection import Collection

from mongorepo import DTO, Access, Index
from mongorepo import exceptions


def _get_collection_and_dto(cls: type, raise_exceptions: bool = True) -> dict[str, Any]:
    """Collect "dto" and "collection" attributes from `Meta` class"""
    attributes: dict[str, Any] = {}
    meta = get_meta(cls)
    try:
        dto = meta.dto
        attributes['dto'] = dto
    except AttributeError as e:
        if raise_exceptions:
            raise exceptions.NoDTOTypeException from e
        else:
            attributes['dto'] = None
    try:
        collection = meta.collection
        attributes['collection'] = collection
    except AttributeError as e:
        if raise_exceptions:
            raise exceptions.NoCollectionException from e
        else:
            attributes['collection'] = None

    return attributes


def _get_meta_attributes(cls, raise_exceptions: bool = True) -> dict[str, Any]:
    """
    Collect all available attributes from `Meta` class
    * use `mongorepo.get_available_meta_attributes` to get all available `Meta` attributes
    """
    attributes: dict[str, Any] = _get_collection_and_dto(
        cls=cls, raise_exceptions=raise_exceptions,
    )
    meta = get_meta(cls)

    index: Index | str | None = getattr(meta, 'index', None)
    attributes['index'] = index

    method_access: Access | None = getattr(meta, 'method_access', None)
    attributes['method_access'] = method_access

    substitute: dict[str, str] | None = getattr(meta, 'substitute', None)
    attributes['substitute'] = substitute

    return attributes


def get_meta(cls: type) -> Any:
    """
    Tries to get `Meta` class for the class or raises an exception
    """
    try:
        meta = cls.__dict__['Meta']
    except (AttributeError, KeyError) as e:
        raise exceptions.NoMetaException from e
    if not isinstance(meta, type):
        raise exceptions.NoMetaException
    return meta


def get_dto_type_hints(dto: Type[DTO] | DTO) -> dict[str, Any]:
    """Returns dictionary of fields' type hints for a dataclass or dataclass instance"""
    default_values = {}
    for field_name, value in dto.__annotations__.items():
        origin = get_origin(value)
        if isinstance(value, type):
            default_values[field_name] = value
        elif origin is list:
            default_values[field_name] = origin
        elif origin is Union:
            args = get_args(value)
            default_values[field_name] = get_origin(args[0])
    return default_values


def create_index(index: Index | str, collection: Collection) -> None:
    """
    ### Creates an index for the collection
    * index parameter can be string or `mongorepo.Index`
    * If index is string, create standard mongodb index
    * If it's `mongorepo.Index` creates index with user's settings
    """
    if isinstance(index, str):
        collection.create_index(index)
        return
    index_name = f'index_{index.field}'
    if index.name:
        index_name = index.name
    direction = pymongo.DESCENDING if index.desc else pymongo.ASCENDING
    collection.create_index(
        [(index.field, direction)],
        name=index_name,
        unique=index.unique,
    )


def get_prefix(access: Access | None, cls: type | None = None) -> str:
    """
    Returns string prefix according to Access value,
    * it can be `'_'`, `'__'`, `_{cls.__name__}__` or `''`
    """
    match access:
        case Access.PRIVATE:
            prefix = f'_{cls.__name__}__' if cls else '__'
        case Access.PROTECTED:
            prefix = '_'
        case Access.PUBLIC | None:
            prefix = ''
    return prefix


def _get_dto_from_origin(cls: type) -> Any:
    """
    Tries to get "dto" from origin of the class or raises an exception
    ```
    class A[UserDTO]: ...
    _get_dto_from_origin(A)
    UserDTO
    ```
    """
    try:
        if not hasattr(cls, '__orig_bases__'):
            raise exceptions.NoDTOTypeException
        dto: type = get_args(cls.__orig_bases__[0])[0]
    except IndexError:
        raise exceptions.NoDTOTypeException
    if dto is DTO:
        raise exceptions.NoDTOTypeException
    if not is_dataclass(dto):
        raise exceptions.NotDataClass

    return dto


def is_immutable(obj: Any) -> bool:
    """Check if object is immutable"""
    try:
        hash(obj)
    except TypeError:
        return False
    return True


def convert_to_dto(dto_type: Type[DTO], dct: dict[str, Any]) -> DTO:
    """Converts document to dto"""
    if '_id' in dto_type.__dict__['__annotations__']:
        return dto_type(**dct)
    dct.pop('_id')
    return dto_type(**dct)



"""
Library for generating repositories for `MongoDB`, use `mongorepo.decorators` or `mongorepo.classes`
Lib also has async support, so you can use the same classes and decorators
located in `mognorepo.asyncio`
"""
from .base import (
    Access,
    DTO,
    Index,
)
from .docs import get_available_meta_attributes, get_available_repository_methods


__all__ = [
    'Access',
    'DTO', 'Index',
    'get_available_meta_attributes',
    'get_available_repository_methods',
]



from typing import Iterable

from mongorepo import exceptions
from mongorepo.base import Access
from mongorepo.utils import (
    _get_dto_from_origin,
    _get_meta_attributes,
    get_dto_type_hints,
    get_prefix,
)
from mongorepo._methods import (
    _update_integer_field_method,
    _update_list_field_method,
    _get_all_method,
    _get_method,
    _delete_method,
    _add_method,
    _update_method,
    _update_field_method,
    _pop_list_method,
)
from mongorepo.asyncio._methods import (
    _update_field_method_async,
    _update_integer_field_method_async,
    _update_list_field_method_async,
    _add_method_async,
    _delete_method_async,
    _get_all_method_async,
    _get_method_async,
    _update_method_async,
    _pop_list_method_async,
)


def _set_array_fields_methods(
    cls: type,
    array_fields: Iterable[str],
    async_methods: bool = False,
    method_access: Access | None = None,
) -> None:
    """
    Set base methods for array fields
    allow `method_access`, `async_method=True` to make this methods async
    """
    attributes = _get_meta_attributes(cls)
    dto, collection = attributes['dto'], attributes['collection']
    prefix = get_prefix(
        access=attributes['method_access'] if not method_access else method_access, cls=cls
    )
    dto_fields = get_dto_type_hints(dto)
    for field in array_fields:
        if field not in dto_fields:
            raise exceptions.MongoRepoException(message=f'{dto} does not have "{field}" attribute')
        if dto_fields[field] is not list:
            raise exceptions.MongoRepoException(message=f'"{field}" is not of type "list"')
        if async_methods:
            append_method = _update_list_field_method_async(
                dto_type=dto, collection=collection, field_name=field, command='$push'
            )
            remove_method = _update_list_field_method_async(
                dto_type=dto, collection=collection, field_name=field, command='$pull',
            )
            pop_method = _pop_list_method_async(
                dto_type=dto, collection=collection, field_name=field,
            )
        else:
            append_method = _update_list_field_method(
                dto_type=dto, collection=collection, field_name=field, command='$push'
            )
            remove_method = _update_list_field_method(
                dto_type=dto, collection=collection, field_name=field, command='$pull',
            )
            pop_method = _pop_list_method(dto_type=dto, collection=collection, field_name=field)

        pop_method.__name__ = f'{prefix}{field}__pop'
        append_method.__name__ = f'{prefix}{field}__append'
        remove_method.__name__ = f'{prefix}{field}__remove'

        setattr(cls, pop_method.__name__, pop_method)
        setattr(cls, append_method.__name__, append_method)
        setattr(cls, remove_method.__name__, remove_method)


def _set_integer_fields_methods(
    cls: type,
    integer_fields: Iterable[str],
    async_methods: bool = False,
    method_access: Access | None = None,
) -> None:
    """
    Set base methods for integer fields,
    allow `method_access`, `async_method=True` to make this methods async
    """
    attributes = _get_meta_attributes(cls)
    dto, collection = attributes['dto'], attributes['collection']
    prefix = get_prefix(
        access=attributes['method_access'] if not method_access else method_access, cls=cls
    )
    dto_fields = get_dto_type_hints(dto)
    for field in integer_fields:
        if field not in dto_fields:
            raise exceptions.MongoRepoException(message=f'{dto} does not have "{field}" attribute')
        if dto_fields[field] is not int:
            raise exceptions.MongoRepoException(message=f'"{field}" is not integer field')
        if async_methods:
            incr_method = _update_integer_field_method_async(
                dto_type=dto, collection=collection, field_name=field, _weight=1,
            )
            decr_method = _update_integer_field_method_async(
                dto_type=dto, collection=collection, field_name=field, _weight=-1,
            )
        else:
            incr_method = _update_integer_field_method(
                dto_type=dto, collection=collection, field_name=field, _weight=1,
            )
            decr_method = _update_integer_field_method(
                dto_type=dto, collection=collection, field_name=field, _weight=-1,
            )

        incr_method.__name__ = f'{prefix}increment_{field}'
        decr_method.__name__ = f'{prefix}decrement_{field}'

        setattr(cls, incr_method.__name__, incr_method)
        setattr(cls, decr_method.__name__, decr_method)


def _set_crud_methods(
    cls: type,
    add: bool = True,
    get: bool = True,
    delete: bool = True,
    update: bool = True,
    get_all: bool = True,
    update_field: bool = False,
    async_methods: bool = False,
    method_access: Access | None = None,
) -> None:
    """
    Set CRUD operations to the class
    allow `method_access`, `async_method=True` to make added methods async
    """
    attributes = _get_meta_attributes(cls, raise_exceptions=False)
    collection = attributes['collection']
    prefix = get_prefix(
        access=attributes['method_access'] if not method_access else method_access, cls=cls
    )

    dto = attributes['dto']
    if dto is None:
        dto = _get_dto_from_origin(cls)

    if add:
        f = _add_method_async if async_methods else _add_method
        setattr(cls, f'{prefix}add', f(dto, collection=collection))  # type: ignore
    if update:
        f = _update_method_async if async_methods else _update_method
        setattr(cls, f'{prefix}update', f(dto, collection=collection))  # type: ignore
    if get:
        f = _get_method_async if async_methods else _get_method
        setattr(cls, f'{prefix}get', f(dto, collection=collection))  # type: ignore
    if get_all:
        f = _get_all_method_async if async_methods else _get_all_method
        setattr(cls, f'{prefix}get_all', f(dto, collection=collection))  # type: ignore
    if delete:
        f = _delete_method_async if async_methods else _delete_method
        setattr(cls, f'{prefix}delete', f(dto, collection=collection))  # type: ignore
    if update_field:
        f = _update_field_method_async if async_methods else _update_field_method
        setattr(cls, f'{prefix}update_field', f(dto, collection=collection))  # type: ignore



from dataclasses import dataclass, field


@dataclass(repr=False, eq=False)
class MongoRepoException(Exception):
    message: str | None = field(kw_only=True, default=None)


@dataclass(repr=False, eq=False)
class NoMetaException(MongoRepoException):

    def __str__(self) -> str:
        if self.message:
            return self.message
        return 'Class does not have "Meta" class inside'


@dataclass(repr=False, eq=False)
class NoDTOTypeException(MongoRepoException):
    with_meta: bool = True

    def __str__(self) -> str:
        if self.message:
            return self.message
        return 'DTO type was not provided' + ' in "Meta" class' if self.with_meta else ''


@dataclass(repr=False, eq=False)
class NoCollectionException(MongoRepoException):
    with_meta: bool = True

    def __str__(self) -> str:
        if self.message:
            return self.message
        return 'Collection was not provided' + ' in "Meta" class' if self.with_meta else ''


class NotDataClass(MongoRepoException):
    def __str__(self) -> str:
        if self.message:
            return self.message
        return 'Provided dto type does not implement dataclass interface'


@dataclass(repr=False, eq=False)
class NoSubstituteException(MongoRepoException):
    with_meta: bool = True

    def __str__(self) -> str:
        if self.message:
            return self.message
        return 'Substitute was not provided' + ' in "Meta" class' if self.with_meta else ''


@dataclass(repr=False, eq=False)
class InvalidMethodNameException(MongoRepoException):
    method_name: str
    available_methods: tuple[str, ...] | None = None

    def __str__(self) -> str:
        if self.message:
            return self.message
        message = f'Method "{self.method_name}" does not appear in mongorepo available methods'
        if self.available_methods:
            message += f'\nAvailable methods: {self.available_methods!r}'
        return message


@dataclass(repr=False, eq=False)
class InvalidMethodNameInSourceClassException(MongoRepoException):
    cls_method_name: str
    source_cls: type

    def __str__(self) -> str:
        if self.message:
            return self.message
        return f'{self.source_cls.__name__} does have "{self.cls_method_name}" method'



from dataclasses import asdict, is_dataclass
from typing import Any, AsyncGenerator, Generic

from motor.motor_asyncio import AsyncIOMotorCollection

from mongorepo import exceptions
from mongorepo import DTO, Index
from mongorepo.utils import _get_dto_from_origin, _get_meta_attributes
from mongorepo.asyncio.utils import _run_asyncio_create_index


class AsyncBasedMongoRepository(Generic[DTO]):
    """
    ## Asynchronous base repository class,
    #### Provide DTO type in type hints, example:
    ```
    class DummyMongoRepository(AsyncBasedMongoRepository[UserDTO]):
        ...
    ```
    #### Extend child classes with various methods:
    ```
    async create(self, dto: DTO) -> DTO
    async get(self, **filters) -> DTO | None
    async get_all(self, **filters) -> Iterable[DTO]
    async update(self, dto: DTO, **filters) -> DTO
    async delete(self, **filters) -> bool
    ```
    """

    def __new__(cls) -> 'AsyncBasedMongoRepository':
        try:
            meta = _get_meta_attributes(cls, raise_exceptions=False)
        except exceptions.NoMetaException:
            return super().__new__(cls)
        index: Index | str | None = meta['index']
        if index is not None:
            collection: AsyncIOMotorCollection | None = meta['collection']
            if collection is None:
                raise exceptions.NoCollectionException(
                    message='Cannot access collection from Meta, to create index'
                )
            cls.collection = collection
            _run_asyncio_create_index(index, collection=collection)

        cls.dto_type = _get_dto_from_origin(cls)
        if not is_dataclass(cls.dto_type):
            raise exceptions.NotDataClass

        return super().__new__(cls)

    def __init__(self, collection: AsyncIOMotorCollection | None = None) -> None:
        self.collection = self.__get_collection(collection)
        self.dto_type = _get_dto_from_origin(self.__class__)

    @classmethod
    def __get_collection(cls, collection: AsyncIOMotorCollection | None) -> AsyncIOMotorCollection:
        if collection is not None:
            return collection
        try:
            attrs = _get_meta_attributes(cls, raise_exceptions=False)
        except exceptions.NoMetaException:
            raise exceptions.MongoRepoException(
                message='"Meta" class with "collection" variable was not defined in the class'
            )
        if attrs['collection'] is None:
            raise exceptions.NoCollectionException
        defined_collection = attrs['collection']
        return defined_collection

    def _convert_to_dto(self, dct: dict[str, Any]) -> DTO:
        if hasattr(self.dto_type, '_id'):
            return self.dto_type(**dct)
        dct.pop('_id')
        return self.dto_type(**dct)

    async def get(self, **filters: Any) -> DTO | None:
        result = await self.collection.find_one(filters)
        return self._convert_to_dto(result) if result else None

    async def get_all(self, **filters: Any) -> AsyncGenerator[DTO, None]:
        cursor = self.collection.find(filters)
        async for doc in cursor:
            yield self._convert_to_dto(doc)

    async def update(self, dto: DTO, **filters: Any) -> DTO:
        data: dict[str, dict[str, Any]] = {'$set': {}}
        for field, value in asdict(dto).items():
            if isinstance(value, (int, bool)):
                data['$set'][field] = value
            elif not field:
                continue
            data['$set'][field] = value
        await self.collection.find_one_and_update(filter=filters, update=data)
        return dto

    async def delete(self, **filters: Any) -> bool:
        deleted = await self.collection.find_one_and_delete(filters)
        return True if deleted else False

    async def add(self, dto: DTO) -> DTO:
        await self.collection.insert_one(asdict(dto))
        return dto



"""## !Package not implemented"""
from typing import Any, Callable, Type, TypeVar

from pymongo.collection import Collection

from mongorepo import exceptions
from mongorepo import DTO
from mongorepo.utils import _get_meta_attributes
from mongorepo._methods import METHOD_NAME__CALLABLE


def substitute_class_methods(target_cls: type, source_cls: type) -> type:
    attributes = _get_meta_attributes(target_cls)
    substitute: dict[str, str] | None = attributes['substitute']
    dto, collection = attributes['dto'], attributes['collection']

    if not substitute:
        raise exceptions.NoSubstituteException

    for mongo_method, cls_method in substitute.items():
        __substitute_method(
            method_name=mongo_method,
            target_cls=target_cls,
            source_cls=source_cls,
            cls_method=cls_method,
            dto=dto,
            collection=collection,
        )
    return target_cls


def __substitute_method(
    method_name: str,
    target_cls: type,
    source_cls: type,
    cls_method: str,
    dto: Type[DTO],
    collection: Collection,
) -> None:
    _method = METHOD_NAME__CALLABLE.get(method_name, None)
    if not _method:
        raise exceptions.InvalidMethodNameException(
            method_name,
            available_methods=tuple(METHOD_NAME__CALLABLE.keys()),
        )

    _cls_method: Callable | None = getattr(source_cls, cls_method, None)
    if not _cls_method:
        raise exceptions.InvalidMethodNameInSourceClassException(cls_method, source_cls)

    get_: Callable = _method(dto_type=dto, collection=collection)

    get_.__annotations__ = _cls_method.__annotations__
    get_.__name__ = _cls_method.__name__
    print(get_.__annotations__)

    replace_typevar_in_annotations(dto, get_)

    setattr(target_cls, cls_method, get_)


def replace_typevar_in_annotations(type_name: Any, func: Callable) -> None:
    for hint, val in func.__annotations__.items():
        if isinstance(val, TypeVar):
            func.__annotations__[hint] = type_name





from mongorepo.asyncio.utils import _run_asyncio_create_index
from mongorepo.base import Access
from mongorepo.setters import (
    _set_array_fields_methods,
    _set_crud_methods,
    _set_integer_fields_methods,
)
from mongorepo.utils import _get_meta_attributes, create_index


def _handle_cls(
    cls,
    add: bool,
    get: bool,
    get_all: bool,
    update: bool,
    delete: bool,
    update_field: bool,
    integer_fields: list[str] | None,
    array_fields: list[str] | None,
    method_access: Access | None,
) -> type:
    """
    Calls for functions that set different methods and attributes to the class
    """
    attributes = _get_meta_attributes(cls, raise_exceptions=False)
    collection = attributes['collection']
    index = attributes['index']

    _set_crud_methods(
        cls,
        add=add,
        get=get,
        get_all=get_all,
        update=update,
        delete=delete,
        update_field=update_field,
        method_access=method_access,
    )

    if index is not None:
        create_index(index=index, collection=collection)

    if integer_fields is not None:
        _set_integer_fields_methods(cls, integer_fields=integer_fields, method_access=method_access)

    if array_fields is not None:
        _set_array_fields_methods(cls, array_fields=array_fields, method_access=method_access)

    return cls


def _handle_cls_async(
    cls,
    add: bool,
    get: bool,
    get_all: bool,
    update: bool,
    delete: bool,
    update_field: bool,
    integer_fields: list[str] | None,
    array_fields: list[str] | None,
    method_access: Access | None,
) -> type:
    """
    Calls for functions that set different async methods and attributes to the class
    """
    attributes = _get_meta_attributes(cls, raise_exceptions=False)
    collection = attributes['collection']
    index = attributes['index']

    _set_crud_methods(
        cls,
        add=add,
        get=get,
        get_all=get_all,
        update=update,
        delete=delete,
        update_field=update_field,
        method_access=method_access,
        async_methods=True,
    )

    if integer_fields is not None:
        _set_integer_fields_methods(
            cls, integer_fields=integer_fields, async_methods=True, method_access=method_access,
        )

    if array_fields is not None:
        _set_array_fields_methods(
            cls, array_fields=array_fields, async_methods=True, method_access=method_access,
        )

    if index is not None:
        _run_asyncio_create_index(index=index, collection=collection)
    return cls
